#include "data/package_line.h"
#include "data/data_coordinates.h"
#include "data_manager/data_manager.h"

namespace slam
{

PackageLine::PackageLine()
{
}

std::vector<slam::Vector3d> calculateCurvatures(const std::vector<slam::Vector2d>& points) {
    std::vector<slam::Vector3d> curvatures;
    if (points.size() < 3) {
        std::cerr << "Need at least 3 points to calculate curvatures" << std::endl;
        return curvatures; // Return empty vector as curvature calculation is not possible
    }

    // Insert zero curvature for the first point
    curvatures.push_back({points[0].x, points[0].y, 0});

    for (size_t i = 1; i < points.size() - 1; ++i) {
        const auto& p1 = points[i - 1];
        const auto& p2 = points[i];
        const auto& p3 = points[i + 1];

        double num = std::abs((p2.x - p1.x) * (p3.y - p2.y) - (p2.y - p1.y) * (p3.x - p2.x));
        double den = std::sqrt(
            (std::pow(p3.x - p1.x, 2) + std::pow(p3.y - p1.y, 2)) *
            (std::pow(p2.x - p1.x, 2) + std::pow(p2.y - p1.y, 2)) *
            (std::pow(p3.x - p2.x, 2) + std::pow(p3.y - p2.y, 2))
        );

        double curvature = 0;
        if (den != 0) {
            curvature = 2 * num / den;
        }

        // Store the point with its curvature as z value
        curvatures.push_back({p2.x, p2.y, curvature});
    }

    // Insert zero curvature for the last point
    curvatures.push_back({points.back().x, points.back().y, 0});

    return curvatures;
}

void PackageLine::Init()
{
    start_point_ = GetVectorPoint(coordinates[0].latitude, coordinates[0].longitude);

    slam::Vector2d last_point_ = start_point_;

    for (size_t i = 0; i < COOR_P_NUM; i++)
    {
        slam::Vector2d c_world_p = GetVectorPoint(coordinates[i].latitude, coordinates[i].longitude);
        c_world_p.x -= start_point_.x;
        c_world_p.y -= start_point_.y;

        if(i>0)
        {
            double td_ =  fabs(c_world_p.x - last_point_.x) +fabs(c_world_p.y - last_point_.y);
            if(td_ < 0.5)
            {
                continue;
            }else
            {
                last_point_ = c_world_p;
            }
        }

        ref_line_.push_back(c_world_p);
    }

    ref_line_curve_ = calculateCurvatures(ref_line_);

    if(!ref_line_.empty())cur_kd_tree_ = SlamUtil::buildKDTree(ref_line_);
}

std::vector<slam::Vector3d> extractPoints(const std::vector<slam::Vector3d> &ref_line, int index, int forward_num, int backward_num)
{
    std::vector<slam::Vector3d> extracted_points;

    // 计算起始和结束索引
    int start_index = std::max(index - forward_num, 0);
    int end_index = std::min(index + backward_num, static_cast<int>(ref_line.size()) - 1);

    extracted_points.resize(end_index - start_index + 1);

    int cindex = 0;
    // 提取点
    for (int i = start_index; i <= end_index; ++i)
    {
        extracted_points[cindex] = ref_line[i];
        cindex++;
    }

    return extracted_points;
}

std::vector<slam::Vector3d> extractPointsFromLoop(const std::vector<slam::Vector3d> &ref_line, int index, int forward_num, int backward_num)
{
    std::vector<slam::Vector3d> extracted_points;
    int n = ref_line.size();

    // 调整提取的总点数
    extracted_points.reserve(forward_num + backward_num + 1);

    // 向前提取点，环状循环
    for (int i = index - forward_num; i <= index + backward_num; ++i) {
        int adjusted_index = (i + n) % n;  // 正确处理负索引和超出范围的索引
        extracted_points.push_back(ref_line[adjusted_index]);
    }

    return extracted_points;
}

bool PackageLine::PackageOdom(slam::slam_odometry & odom_)
{
    odom_.utc_timestamp = cur_ins_.utc_timestamp;
    odom_.position.position.x = cur_host_pos.x;
    odom_.position.position.y = cur_host_pos.y;

    Quaterniond wk_quat_ret = SlamUtil::EulerToQuaternion(
                cur_ins_.attitude.z * DEG2RAD,
                cur_ins_.attitude.y * DEG2RAD,
                cur_ins_.attitude.x * DEG2RAD);

    odom_.position.orientation.w = (float)wk_quat_ret.w;
    odom_.position.orientation.x = (float)wk_quat_ret.x;
    odom_.position.orientation.y = (float)wk_quat_ret.y;
    odom_.position.orientation.z = (float)wk_quat_ret.z;

    return true;
}

bool PackageLine::PackageRM(std::vector<line_point>& points)
{
    //int c_index = SlamUtil::findNearestPointIndex(cur_host_pos, ref_line_);
    double out_min_dist;

    if (cur_kd_tree_ == nullptr)return false;
    int c_index =  SlamUtil::findNearest(cur_kd_tree_, cur_host_pos, out_min_dist);
    if (c_index == -1)return false;
    if(out_min_dist > 5)return false;

    std::cout << "slam select point index:" << c_index << std::endl;
    if(extract_loop_flag)
    {
        c_line = extractPointsFromLoop(ref_line_curve_, c_index, 20, 180);
    }else
    {
        c_line = extractPoints(ref_line_curve_, c_index, 20, 180);
    }


    if(points.size() > c_line.size())points.resize(c_line.size());

    for (uint i = 0 ; i< points.size(); ++i)
    {
        slam::Vector2d cp;
        cp.x = c_line[i].x;
        cp.y = c_line[i].y;

        slam::Vector2d wk_vcs_p = SlamUtil::globalToLocal(cp, cur_host_pos, cur_host_heading);

        points[i].point.x = wk_vcs_p.x;
        points[i].point.y = wk_vcs_p.y;
        points[i].point_attri.curve = c_line[i].z;
    }

    return true;
}

void PackageLine::SetHostVehiclePosition(const InsData& _ins)
{
    if (last_odom_utc_ >= _ins.utc_timestamp) return;
    last_odom_utc_ =  _ins.utc_timestamp;

    cur_host_pos = GetVectorPoint(_ins.latitude, _ins.longitude);
    cur_host_pos.x -= start_point_.x;
    cur_host_pos.y -= start_point_.y;

    cur_host_heading = _ins.heading * DEG2RAD;

    cur_ins_ = _ins;
}

slam::Vector2d PackageLine::GetVectorPoint(double lat, double lon)
{
    slam::Vector2d wk_p0;

    Localization::PointLoc::CoordinateUtil::PositionLL wk_gps;
    wk_gps.latitude = lat;
    wk_gps.longitude = lon;
    Localization::PointLoc::CoordinateUtil::PositionGK world_p_ = Localization::PointLoc::CoordinateUtil::Ll2GK(wk_gps, base_lon_);

    wk_p0.x = world_p_.x;
    wk_p0.y = world_p_.y;

    return wk_p0;
}

}
